package de.mdelab.mlsdm.interpreter.facade;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

public class OptimizedMSLDMInstanceFacade extends MLSDMInstanceFacade {

	/*
	 * This map use used for reverse navigation of unidirectional links. It maps
	 * the link type to a mapping between link target objects and a list of all
	 * source objects pointing to that target object with the specified
	 * reference.
	 */
	private final Map<EReference, Map<EObject, Set<EObject>>> inverseLinks;
	
	private SDMChangeListener changeListener;
	
	private final Collection<EReference> unidirectionalReferences;
	
	public OptimizedMSLDMInstanceFacade( final Collection<EReference> p_unidirectionalReferences  ) {
		inverseLinks = new HashMap<EReference, Map<EObject, Set<EObject>>>();
		unidirectionalReferences = p_unidirectionalReferences;

		changeListener = null;//new SDMChangeListener( this );
	}

	@Override
	public void delete(Object p_object) {
		final EObject objToDelete = (EObject) p_object;

		/*
		 * Maintain inverse links map
		 */
		for (final Entry<EReference, Map<EObject, Set<EObject>>> entry : inverseLinks.entrySet()) {
			final EReference reference = entry.getKey();
			
			assert reference.isChangeable();

			
			final Map<EObject, Set<EObject>> map = entry.getValue();

			final Iterator<Entry<EObject, Set<EObject>>> it = map.entrySet().iterator();

			while (it.hasNext()) {
				final Entry<EObject, Set<EObject>> entry2 = it.next();
				final EObject objKey = entry2.getKey();
				

				if ( p_object == objKey ) {
					for (final EObject src : entry2.getValue())	{
						
						if ( reference.isMany() ) {
							((Collection<?>) src.eGet( reference ) ) .remove( objKey );
						}
						else {
							src.eSet( reference, null );
						}
					}

					it.remove();
				}

				entry2.getValue().remove( p_object );
			}
		}

		/*
		 * Delete the object from the model
		 */
		// EcoreUtil.delete(eObject, true);
		for (final EReference eReference : objToDelete.eClass().getEAllReferences())
		{
			if (eReference.isChangeable())
			{
				if (!eReference.isMany())
				{
					objToDelete.eSet(eReference, null);
				}
				else
				{
					((Collection<?>) objToDelete.eGet(eReference)).clear();
				}
			}
		}

		final EStructuralFeature containingFeature = objToDelete.eContainingFeature();

		if (containingFeature != null)
		{
			assert containingFeature.isChangeable();

			if (!containingFeature.isMany())
			{
				objToDelete.eContainer().eSet(containingFeature, null);
			}
			else
			{
				((Collection<?>) objToDelete.eContainer().eGet(containingFeature)).remove(p_object);
			}
		}
	}

	void unregisterInverseLink(	final EObject linkSource,
								final EObject linkTarget,
								final EReference linkType )	{
		assert linkType.isChangeable();

		final Map<EObject, Set<EObject>> map = this.inverseLinks.get(linkType);

		if (map != null)
		{
			final Set<EObject> set = map.get(linkTarget);

			if (set != null)
			{
				set.remove(linkSource);
			}
		}
	}

	void detachChangeListener() {
		if ( changeListener != null ) {
			changeListener.detachAdapter();
			changeListener = null;
		}
	}

	@SuppressWarnings("unchecked")
	private void storeInverseLinks(final EObject eObject) {
		for (final EReference eReference : eObject.eClass().getEAllReferences()) {
			/*
			 * Only consider changeable, unidirectional, non-containment
			 * references
			 */
			if (eReference.isChangeable() && !eReference.isContainment() && eReference.getEOpposite() == null)
			{
				Map<EObject, Set<EObject>> map1 = this.inverseLinks.get(eReference);

				if (map1 == null)
				{
					map1 = new HashMap<EObject, Set<EObject>>();
					this.inverseLinks.put(eReference, map1);
				}

				Collection<EObject> targetObjects = null;

				if (!eReference.isMany())
				{
					targetObjects = new ArrayList<EObject>();
					final EObject o = (EObject) eObject.eGet(eReference);

					if (o != null)
					{
						targetObjects.add(o);
					}
				}
				else
				{
					targetObjects = (Collection<EObject>) eObject.eGet(eReference);
				}

				for (final EObject targetObject : targetObjects)
				{
					Set<EObject> set = map1.get(targetObject);

					if (set == null)
					{
						set = new HashSet<EObject>();
						map1.put(targetObject, set);
					}

					set.add(eObject);
				}
			}
		}
	}
	
	public void reset() {
		detachChangeListener();
		inverseLinks.clear();
	}

	public void initialize( final Set<EObject> p_models ) {
		reset();
		
		for ( final EObject model : p_models ) {
			storeInverseLinks( model );

			for (final TreeIterator<EObject> it = model.eAllContents(); it.hasNext();) {
				storeInverseLinks( it.next() );
			}
		}

		/*
		 * Add EContentAdapters
		 */
		changeListener = new SDMChangeListener( this );
		changeListener.attachAdapter( p_models );
	}

	EList<EObject> getLinkSources(final EReference linkType, final EObject linkTarget) {
		final Map<EObject, Set<EObject>> map = this.inverseLinks.get(linkType);

		if (map != null)
		{
			final Set<EObject> set = map.get(linkTarget);

			if (set != null)
			{
				return new BasicEList<EObject>(set);
			}
		}

		return new BasicEList<EObject>();
	}

	Collection<EReference> getUnidirectionalReferences() {
		return unidirectionalReferences;
	}

	void registerInverseLink(final EObject linkSource, final EObject linkTarget, final EReference linkType)
	{
		assert linkType.isChangeable();

		Map<EObject, Set<EObject>> map = this.inverseLinks.get(linkType);

		if (map == null)
		{
			map = new HashMap<EObject, Set<EObject>>();
			this.inverseLinks.put(linkType, map);
		}

		Set<EObject> set = map.get(linkTarget);

		if (set == null)
		{
			set = new HashSet<EObject>();
			map.put(linkTarget, set);
		}

		set.add(linkSource);
	}
}
