package de.mdelab.mlsdm.interpreter.facade;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EContentAdapter;

public class SDMChangeListener extends EContentAdapter {

	private boolean							active;

	private final Set<EObject>				rootObjects;
	
	private final OptimizedMSLDMInstanceFacade sdmInstanceFacade;

	public SDMChangeListener( final OptimizedMSLDMInstanceFacade p_sdmInstanceFacade ) {
		super();

		this.active = true;

		this.rootObjects = new HashSet<EObject>();
		
		sdmInstanceFacade = p_sdmInstanceFacade;
	}

	public void activate() {
		assert !this.active;

		this.active = true;
	}

	public void deactivate() {
		assert this.active;

		this.active = false;
	}

	public void attachAdapter(final Collection<EObject> rootObjects) {
		this.rootObjects.addAll(rootObjects);

		for (final EObject eObject : this.rootObjects)
		{
			eObject.eAdapters().add(this);
		}

	}

	public void detachAdapter()
	{
		for (final EObject eObject : this.rootObjects)
		{
			eObject.eAdapters().remove(this);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public void notifyChanged(final Notification notification) {
		super.notifyChanged(notification);

		if (this.active && notification.getNotifier() instanceof EObject && !notification.isTouch()) {
			final EObject notifier = (EObject) notification.getNotifier();

			if (notification.getFeature() instanceof EReference) {
				final EReference eReference = (EReference) notification.getFeature();

				switch (notification.getEventType())
				{
					case Notification.ADD:
						this.elementsAdded(notifier, Collections.singletonList((EObject) notification.getNewValue()), eReference);
						break;

					case Notification.ADD_MANY:
						this.elementsAdded(notifier, (Collection<EObject>) notification.getNewValue(), eReference);
						break;

					case Notification.REMOVE:
						this.elementsRemoved(notifier, Collections.singletonList((EObject) notification.getOldValue()), eReference);
						break;

					case Notification.REMOVE_MANY:
						this.elementsRemoved(notifier, (Collection<EObject>) notification.getOldValue(), eReference);
						break;

					case Notification.SET:
						if (notification.getNewValue() != null)
						{
							this.elementsAdded(notifier, Collections.singletonList((EObject) notification.getNewValue()), eReference);
						}

						if (notification.getOldValue() != null)
						{
							this.elementsRemoved(notifier, Collections.singletonList((EObject) notification.getOldValue()), eReference);
						}
						break;
				}
			}
		}
	}

	private void elementsAdded(final EObject notifier, final Collection<EObject> elements, final EReference reference) {
		/*
		 * Maintain uncoveredElements and inverse links for the elements and
		 * their subtrees
		 */
		if (reference.isContainment())
		{
			final Set<EObject> allElements = new HashSet<EObject>();

			for (final EObject eObject : elements)
			{
				allElements.add(eObject);

				final Iterator<EObject> it = eObject.eAllContents();

				while (it.hasNext())
				{
					allElements.add(it.next());
				}
			}

			for (final EObject eObject : allElements) {
				this.registerAllInverseLinks(eObject);
			}
		}
		else if (reference.getEOpposite() == null)
		{
			for (final EObject eObject : elements) {
				sdmInstanceFacade.registerInverseLink(notifier, eObject, reference);
			}
		}
	}

	private void elementsRemoved(final EObject notifier, final Collection<EObject> elements, final EReference reference)
	{
		/*
		 * Maintain uncoveredElements and inverse links for the elements and
		 * their subtrees
		 */
		if (reference.isContainment())
		{
			final Set<EObject> allElements = new HashSet<EObject>();

			for (final EObject eObject : elements)
			{
				allElements.add(eObject);

				final Iterator<EObject> it = eObject.eAllContents();

				while (it.hasNext())
				{
					allElements.add(it.next());
				}
			}

			for (final EObject eObject : allElements)
			{
				this.unregisterAllInverseLinks(eObject);
			}
		}
		else if (sdmInstanceFacade.getUnidirectionalReferences().contains(reference))
		{
			for (final EObject eObject : elements)
			{
				sdmInstanceFacade.unregisterInverseLink(notifier, eObject, reference);
			}
		}
	}

	@SuppressWarnings("unchecked")
	private void registerAllInverseLinks(final EObject eObject)
	{
		for (final EReference eReference : eObject.eClass().getEAllReferences())
		{
			if (sdmInstanceFacade.getUnidirectionalReferences().contains(eReference))
			{
				if (!eReference.isMany())
				{
					if (eObject.eGet(eReference) != null)
					{
						sdmInstanceFacade.registerInverseLink(eObject, (EObject) eObject.eGet(eReference), eReference);
					}
				}
				else
				{
					for (final EObject referencedEObject : (Collection<EObject>) eObject.eGet(eReference))
					{
						sdmInstanceFacade.registerInverseLink(eObject, referencedEObject, eReference);
					}
				}
			}
		}
	}

	@SuppressWarnings("unchecked")
	private void unregisterAllInverseLinks(final EObject eObject)
	{
		for (final EReference eReference : eObject.eClass().getEAllReferences())
		{
			if (sdmInstanceFacade.getUnidirectionalReferences().contains(eReference))
			{
				if (!eReference.isMany())
				{
					if (eObject.eGet(eReference) != null)
					{
						sdmInstanceFacade.unregisterInverseLink(eObject, (EObject) eObject.eGet(eReference), eReference);
					}
				}
				else
				{
					for (final EObject referencedEObject : (Collection<EObject>) eObject.eGet(eReference))
					{
						sdmInstanceFacade.unregisterInverseLink(eObject, referencedEObject, eReference);
					}
				}
			}
		}
	}
}
